import argparse
from pprint import pprint

import pytest

from nucs.examples.queens_problem import QueensProblem
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.heuristics import min_value_dom_heuristic, smallest_domain_var_heuristic
from nucs.statistics import STATS_SOLVER_SOLUTION_NB, get_statistics


class TestQueens:
    @pytest.mark.parametrize(
        "queen_nb,solution_nb", [(1, 1), (2, 0), (3, 0), (4, 2), (5, 10), (6, 4), (7, 40), (8, 92), (9, 352)]
    )
    def test_queens_solve(self, queen_nb: int, solution_nb: int) -> None:
        problem = QueensProblem(queen_nb)
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_SOLVER_SOLUTION_NB] == solution_nb

    def test_queens_8_solve_ff(self) -> None:
        problem = QueensProblem(8)
        solver = BacktrackSolver(
            problem, var_heuristic=smallest_domain_var_heuristic, dom_heuristic=min_value_dom_heuristic
        )
        solver.solve_all()
        assert solver.statistics[STATS_SOLVER_SOLUTION_NB] == 92


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-n", type=int, default=12)
    args = parser.parse_args()
    problem = QueensProblem(args.n)
    solver = BacktrackSolver(problem)
    solver.solve_all()
    pprint(get_statistics(solver.statistics))



import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import MAX, MIN, PROP_CONSISTENCY, PROP_INCONSISTENCY
from nucs.numpy import new_triggers


def get_complexity_max_eq(n: int, data: NDArray) -> float:
    return 3 * n


def get_triggers_max_eq(n: int, data: NDArray) -> NDArray:
    """
    Returns the triggers for this propagator.
    :param n: the number of variables
    :return: an array of triggers
    """
    triggers = new_triggers(n, True)
    triggers[-1, MIN] = False
    return triggers


@njit(cache=True)
def compute_domains_max_eq(domains: NDArray, data: NDArray) -> int:
    """
    Implements Max_i x_i = x_{n-1}.
    :param domains: the domains of the variables
    :param data: unused here
    """
    x = domains[:-1]
    y = domains[-1]
    y[MIN] = max(y[MIN], np.max(x[:, MIN]))
    y[MAX] = min(y[MAX], np.max(x[:, MAX]))
    if y[MIN] > y[MAX]:
        return PROP_INCONSISTENCY
    for i in range(len(x)):
        x[i, MAX] = min(x[i, MAX], y[MAX])
        if x[i, MAX] < x[i, MIN]:
            return PROP_INCONSISTENCY
    return PROP_CONSISTENCY



import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import MAX, MIN, PROP_CONSISTENCY, PROP_INCONSISTENCY
from nucs.numpy import new_triggers


def get_complexity_affine_eq(n: int, data: NDArray) -> float:
    return 5 * n


def get_triggers_affine_eq(n: int, data: NDArray) -> NDArray:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param n: the number of variables
    :return: an array of triggers
    """
    return new_triggers(n, True)


@njit(cache=True)
def compute_domain_sum_min(domains: NDArray, data: NDArray) -> int:
    domain_sum_min = data[-1]
    for i, c in enumerate(data[:-1]):
        domain_sum_min -= c * (domains[i, MAX] if c > 0 else domains[i, MIN])
    return domain_sum_min


@njit(cache=True)
def compute_domain_sum_max(domains: NDArray, data: NDArray) -> int:
    domain_sum_max = data[-1]
    for i, c in enumerate(data[:-1]):
        domain_sum_max -= c * (domains[i, MIN] if c > 0 else domains[i, MAX])
    return domain_sum_max


@njit(cache=True)
def compute_domains_affine_eq(domains: NDArray, a: NDArray) -> int:
    """
    Implements Sigma_i a_i * x_i = a_{n-1}.
    :param domains: the domains of the variables
    :param a: the parameters of the propagator
    """
    domain_sum_min = compute_domain_sum_min(domains, a)
    domain_sum_max = compute_domain_sum_max(domains, a)
    new_domains = np.copy(domains)
    for i, c in enumerate(a[:-1]):
        if c != 0:
            if c > 0:
                new_min = domains[i, MAX] - (domain_sum_min // -c)
                new_max = domains[i, MIN] + (domain_sum_max // c)
            else:
                new_min = domains[i, MAX] - (-domain_sum_max // c)
                new_max = domains[i, MIN] + (-domain_sum_min // -c)
            new_domains[i, MIN] = max(domains[i, MIN], new_min)
            new_domains[i, MAX] = min(domains[i, MAX], new_max)
            if new_domains[i, MIN] > new_domains[i, MAX]:
                return PROP_INCONSISTENCY
    domains[:] = new_domains[:]  # numba does not support copyto
    return PROP_CONSISTENCY



from nucs.constants import MAX, MIN
from nucs.problems.problem import Problem, is_solved


class TestProblem:
    def test_set_min_value(self) -> None:
        problem = Problem([(0, 2), (0, 2), (0, 2)], [0, 1, 2], [1, 1, 1])
        problem.init_problem()
        problem.set_min_value(1, 2)
        assert problem.shr_domains_arr[1, MIN] == 1
        assert problem.shr_domains_arr[1, MAX] == 2

    def test_is_not_solved_ok(self) -> None:
        problem = Problem([(0, 2), (0, 2), (4, 6)])
        problem.init_problem()
        assert not is_solved(problem.shr_domains_arr)

    def test_is_not_solved_ko(self) -> None:
        problem = Problem([(2, 2), (2, 2), (6, 6)])
        problem.init_problem()
        assert is_solved(problem.shr_domains_arr)



from pprint import pprint

from nucs.examples.bibd_problem import BIBDProblem
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.statistics import STATS_SOLVER_SOLUTION_NB, get_statistics


class TestBIBD:
    def test_6_10_5_3_2(self) -> None:
        problem = BIBDProblem(6, 10, 5, 3, 2)
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_SOLVER_SOLUTION_NB] == 1

    def test_8_14_7_4_3(self) -> None:
        problem = BIBDProblem(8, 14, 7, 4, 3)
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_SOLVER_SOLUTION_NB] == 92


if __name__ == "__main__":
    problem = BIBDProblem(8, 14, 7, 4, 3)
    solver = BacktrackSolver(problem)
    solver.solve_all()
    pprint(get_statistics(solver.statistics))



from typing import List

from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_AFFINE_EQ, ALG_AFFINE_LEQ


class KnapsackProblem(Problem):
    """
    CSPLIB problem #133 - https://www.csplib.org/Problems/prob133/
    """

    def __init__(self, weights: List[int], volumes: List[int], capacity: int) -> None:
        n = len(weights)
        super().__init__([(0, 1)] * n + [(0, sum(weights))])
        self.add_propagator((list(range(n)), ALG_AFFINE_LEQ, [*volumes, capacity]))
        self.add_propagator((list(range(n + 1)), ALG_AFFINE_EQ, [*weights, -1, 0]))
        self.weight = n



import argparse
from pprint import pprint

import pytest

from nucs.constants import MIN
from nucs.examples.golomb_problem import GolombProblem, golomb_consistency_algorithm, index, init_domains
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.statistics import get_statistics


class TestGolomb:

    @pytest.mark.parametrize(
        "mark_nb,i,j,idx", [(4, 0, 1, 0), (4, 0, 2, 1), (4, 0, 3, 2), (4, 1, 2, 3), (4, 1, 3, 4), (4, 2, 3, 5)]
    )
    def test_index(self, mark_nb: int, i: int, j: int, idx: int) -> None:
        assert index(mark_nb, i, j) == idx

    def test_init_domains(self) -> None:
        domains = init_domains(6, 4)
        assert domains[:, MIN].tolist() == [1, 3, 6, 1, 3, 1]

    @pytest.mark.parametrize("mark_nb,solution_nb", [(4, 6), (5, 11), (6, 17), (7, 25), (8, 34), (9, 44)])
    def test_golomb(self, mark_nb: int, solution_nb: int) -> None:
        problem = GolombProblem(mark_nb)
        solver = BacktrackSolver(problem, consistency_algorithm=golomb_consistency_algorithm)
        solution = solver.minimize(problem.length_idx)
        assert solution
        assert solution[problem.length_idx] == solution_nb


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-n", type=int, default=10)
    args = parser.parse_args()
    problem = GolombProblem(args.n)
    solver = BacktrackSolver(problem, consistency_algorithm=golomb_consistency_algorithm)
    solution = solver.minimize(problem.length_idx)
    pprint(get_statistics(solver.statistics))
    print(solution)
    print(solution[problem.length_idx])  # type: ignore



from typing import Callable, Iterator, List, Optional

import numpy as np

from nucs.constants import MIN, PROBLEM_INCONSISTENT, PROBLEM_SOLVED
from nucs.problems.problem import Problem
from nucs.solvers.consistency_algorithms import bound_consistency_algorithm
from nucs.solvers.heuristics import first_not_instantiated_var_heuristic, min_value_dom_heuristic
from nucs.solvers.solver import Solver
from nucs.statistics import (
    STATS_OPTIMIZER_SOLUTION_NB,
    STATS_SOLVER_BACKTRACK_NB,
    STATS_SOLVER_CHOICE_DEPTH,
    STATS_SOLVER_CHOICE_NB,
    STATS_SOLVER_SOLUTION_NB,
)


class BacktrackSolver(Solver):
    """
    A solver relying on a backtracking mechanism.
    """

    def __init__(
        self,
        problem: Problem,
        consistency_algorithm: Callable = bound_consistency_algorithm,
        var_heuristic: Callable = first_not_instantiated_var_heuristic,
        dom_heuristic: Callable = min_value_dom_heuristic,
    ):
        super().__init__(problem)
        self.choice_points = []  # type: ignore
        self.consistency_algorithm = consistency_algorithm
        self.var_heuristic = var_heuristic
        self.dom_heuristic = dom_heuristic

    def solve(self) -> Iterator[List[int]]:
        while (solution := self.solve_one()) is not None:
            yield solution
            if not self.backtrack():
                break

    def solve_one(self) -> Optional[List[int]]:
        """
        Find at most one solution.
        :return: the solution if it exists or None
        """
        if not self.problem.ready:
            self.problem.init_problem(self.statistics)
            self.problem.ready = True
        while True:
            while (status := self.consistency_algorithm(self.statistics, self.problem)) == PROBLEM_INCONSISTENT:
                if not self.backtrack():
                    return None
            if status == PROBLEM_SOLVED:
                self.statistics[STATS_SOLVER_SOLUTION_NB] += 1
                values = self.problem.shr_domains_arr[self.problem.dom_indices_arr, MIN] + self.problem.dom_offsets_arr
                return values.tolist()
            dom_idx = self.var_heuristic(self.problem.shr_domains_arr)
            shr_domains_copy = self.problem.shr_domains_arr.copy(order="F")
            self.choice_points.append(shr_domains_copy)
            event = self.dom_heuristic(self.problem.shr_domains_arr[dom_idx], shr_domains_copy[dom_idx])
            np.logical_or(
                self.problem.triggered_propagators,
                self.problem.shr_domains_propagators[dom_idx, event],
                self.problem.triggered_propagators,
            )
            self.statistics[STATS_SOLVER_CHOICE_NB] += 1
            cp_max_depth = len(self.choice_points)
            if cp_max_depth > self.statistics[STATS_SOLVER_CHOICE_DEPTH]:
                self.statistics[STATS_SOLVER_CHOICE_DEPTH] = cp_max_depth

    def minimize(self, variable_idx: int) -> Optional[List[int]]:
        solution = None
        while (new_solution := self.solve_one()) is not None:
            solution = new_solution
            self.statistics[STATS_OPTIMIZER_SOLUTION_NB] += 1
            self.reset()
            self.problem.set_max_value(variable_idx, solution[variable_idx] - 1)
        return solution

    def maximize(self, variable_idx: int) -> Optional[List[int]]:
        solution = None
        while (new_solution := self.solve_one()) is not None:
            solution = new_solution
            self.statistics[STATS_OPTIMIZER_SOLUTION_NB] += 1
            self.reset()
            self.problem.set_min_value(variable_idx, solution[variable_idx] + 1)
        return solution

    def backtrack(self) -> bool:
        """
        Backtracks and updates the problem's domains
        :return: true iff it is possible to backtrack
        """
        if len(self.choice_points) == 0:
            return False
        self.statistics[STATS_SOLVER_BACKTRACK_NB] += 1
        self.problem.reset(self.choice_points.pop())  # TODO: optimize by reusing
        return True

    def reset(self) -> None:
        """
        Resets the solver by resetting the problem and the choice points.
        """
        self.choice_points.clear()
        self.problem.reset()



from pprint import pprint

from nucs.examples.alpha_problem import AlphaProblem
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.heuristics import min_value_dom_heuristic, smallest_domain_var_heuristic
from nucs.statistics import STATS_SOLVER_SOLUTION_NB, get_statistics


class TestAlpha:
    def test_alpha(self) -> None:
        problem = AlphaProblem()
        solver = BacktrackSolver(
            problem, var_heuristic=smallest_domain_var_heuristic, dom_heuristic=min_value_dom_heuristic
        )
        solutions = solver.find_all()
        assert solver.statistics[STATS_SOLVER_SOLUTION_NB] == 1
        assert solutions[0][:26] == [
            5,
            13,
            9,
            16,
            20,
            4,
            24,
            21,
            25,
            17,
            23,
            2,
            8,
            12,
            10,
            19,
            7,
            11,
            15,
            3,
            1,
            26,
            6,
            22,
            14,
            18,
        ]


if __name__ == "__main__":
    problem = AlphaProblem()
    solver = BacktrackSolver(
        problem, var_heuristic=smallest_domain_var_heuristic, dom_heuristic=min_value_dom_heuristic
    )
    solutions = solver.find_all()
    pprint(get_statistics(solver.statistics))
    print(solutions[0])



from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import MAX, MIN, PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY
from nucs.numpy import new_triggers


def get_complexity_relation(n: int, data: NDArray) -> float:
    return 3 * len(data)


def get_triggers_relation(n: int, data: NDArray) -> NDArray:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param n: the number of variables
    :return: an array of triggers
    """
    return new_triggers(n, True)


@njit(cache=True)
def compute_domains_relation(domains: NDArray, data: NDArray) -> int:
    """
    :param domains: the domains of the variables
    :param data: the parameters of the propagator
    """
    n = len(domains)
    tuples = data.copy().reshape((-1, n))
    for domain_idx in range(n):
        tuples = tuples[
            (tuples[:, domain_idx] >= domains[domain_idx, MIN]) & (tuples[:, domain_idx] <= domains[domain_idx, MAX])
        ]
        if len(tuples) == 0:
            return PROP_INCONSISTENCY
    for domain_idx in range(n):  # no support for .min(axis=0) in Numba
        domains[domain_idx, MIN] = tuples[:, domain_idx].min()
        domains[domain_idx, MAX] = tuples[:, domain_idx].max()
    if len(tuples) == 1:
        return PROP_ENTAILMENT
    return PROP_CONSISTENCY



import sys

from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import MAX, MIN, PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY
from nucs.numpy import new_triggers


def get_complexity_element_liv(n: int, data: NDArray) -> float:
    return n


def get_triggers_element_liv(n: int, data: NDArray) -> NDArray:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param n: the number of variables
    :return: an array of triggers
    """
    return new_triggers(n, True)


@njit(cache=True)
def compute_domains_element_liv(domains: NDArray, data: NDArray) -> int:
    """
    Enforces l_i = x.
    :param domains: the domains of the variables
    :param data: the parameters of the propagator
    """
    l = domains[:-2]
    i = domains[-2]
    i[MIN] = max(i[MIN], 0)
    i[MAX] = min(i[MAX], len(l) - 1)
    if i[MAX] < i[MIN]:
        return PROP_INCONSISTENCY
    x = domains[-1]
    x_min = sys.maxsize
    x_max = -sys.maxsize
    i_min = i[MIN]
    i_max = i[MAX]
    for idx in range(i[MIN], i[MAX] + 1):
        if x[MAX] < l[idx, MIN] or x[MIN] > l[idx, MAX]:  # no intersection
            if idx == i_min:
                i_min += 1
            elif idx == i_max:
                i_max -= 1
        else:  # intersection
            if l[idx, MIN] < x_min:
                x_min = l[idx, MIN]
            if l[idx, MAX] > x_max:
                x_max = l[idx, MAX]
    if i_max < i_min:
        return PROP_INCONSISTENCY
    i[MIN] = i_min
    i[MAX] = i_max
    x[MIN] = max(x[MIN], x_min)
    x[MAX] = min(x[MAX], x_max)
    if i_min == i_max:
        l[i_min, MIN] = max(l[i_min, MIN], x[MIN])
        l[i_max, MAX] = min(l[i_max, MAX], x[MAX])
        if x[MIN] == x[MAX]:
            return PROP_ENTAILMENT
    return PROP_CONSISTENCY



import pytest

from nucs.problems.latin_square_problem import LatinSquareProblem, LatinSquareRCProblem
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.statistics import STATS_SOLVER_SOLUTION_NB


class TestLatinSquare:
    @pytest.mark.parametrize(
        "size, solution_nb",
        [
            (1, 1),
            (2, 2),
            (3, 12),
            (4, 576),
            # (5, 161280)
        ],
    )
    def test_latin_square(self, size: int, solution_nb: int) -> None:
        problem = LatinSquareProblem(list(range(size)))
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_SOLVER_SOLUTION_NB] == solution_nb

    @pytest.mark.parametrize(
        "size, solution_nb",
        [
            (1, 1),
            (2, 2),
            (3, 12),
            (4, 576),
            # (5, 161280)
        ],
    )
    def test_latin_square_rc(self, size: int, solution_nb: int) -> None:
        problem = LatinSquareRCProblem(size)
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_SOLVER_SOLUTION_NB] == solution_nb



import numpy as np

from nucs.constants import PROP_CONSISTENCY, PROP_ENTAILMENT
from nucs.numpy import new_data_by_values, new_shr_domains_by_values
from nucs.propagators.affine_geq_propagator import compute_domains_affine_geq, get_triggers_affine_geq


class TestAffineGEQ:
    def test_get_triggers(self) -> None:
        data = new_data_by_values([1, -1, 8])
        assert np.all(get_triggers_affine_geq(2, data) == np.array([[False, True], [True, False]]))

    def test_compute_domains_1(self) -> None:
        domains = new_shr_domains_by_values([(1, 10), (1, 10)])
        data = new_data_by_values([1, -1, 1])
        assert compute_domains_affine_geq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[2, 10], [1, 9]]))

    def test_compute_domains_2(self) -> None:
        domains = new_shr_domains_by_values([(5, 10), (5, 10), (5, 10)])
        data = new_data_by_values([1, 1, 1, 27])
        assert compute_domains_affine_geq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[7, 10], [7, 10], [7, 10]]))

    def test_compute_domains_3(self) -> None:
        domains = new_shr_domains_by_values([(5, 10), (1, 2)])
        data = new_data_by_values([1, 1, 6])
        assert compute_domains_affine_geq(domains, data) == PROP_ENTAILMENT
        assert np.all(domains == np.array([[5, 10], [1, 2]]))



from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import PROP_CONSISTENCY
from nucs.numpy import new_triggers


def get_complexity_dummy(n: int, data: NDArray) -> float:
    return 0


def get_triggers_dummy(n: int, data: NDArray) -> NDArray:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param n: the number of variables
    :return: an array of triggers
    """
    return new_triggers(n, True)


@njit(cache=True)
def compute_domains_dummy(domains: NDArray, data: NDArray) -> int:
    """
    A propagator that does nothing.
    """
    return PROP_CONSISTENCY



import numpy as np
from numpy.typing import NDArray

from nucs.constants import MAX, MIN
from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_AFFINE_EQ, ALG_AFFINE_LEQ, ALG_ALLDIFFERENT
from nucs.solvers.consistency_algorithms import bound_consistency_algorithm
from nucs.solvers.heuristics import first_not_instantiated_var_heuristic

GOLOMB_LENGTHS = [0, 0, 1, 3, 6, 11, 17, 25, 34, 44, 55, 72, 85, 106, 127]


def sum_first(n: int) -> int:
    return (n * (n + 1)) // 2


def index(mark_nb: int, i: int, j: int) -> int:
    return i * mark_nb - sum_first(i) + j - i - 1


def init_domains(dist_nb: int, mark_nb: int) -> NDArray:
    domains = np.empty((dist_nb, 2), dtype=np.int32, order="F")
    for i in range(0, mark_nb - 1):
        for j in range(i + 1, mark_nb):
            domains[index(mark_nb, i, j), MIN] = GOLOMB_LENGTHS[j - i + 1] if j - i + 1 < mark_nb else sum_first(j - i)
    domains[:, MAX] = sum_first(dist_nb)
    return domains


class GolombProblem(Problem):
    """
    This is the famous Golomb ruler problem.

    It consists in finding n integers mark_i such that:
    - mark_0 = 0,
    - mark_0 <...< mark_n-1,
    - for all i < j, mark_j - mark_i are different,
    - mark_n-1 is minimal.

    CSPLIB problem #6 - https://www.csplib.org/Problems/prob006/
    """

    def __init__(self, mark_nb: int) -> None:
        self.mark_nb = mark_nb
        self.dist_nb = sum_first(mark_nb - 1)  # this the number of distances
        super().__init__(init_domains(self.dist_nb, mark_nb))
        self.length_idx = index(mark_nb, 0, mark_nb - 1)  # we want to minimize this
        # Additional items used in prune():
        # - a reusable array for storing the minimal sum of different integers:
        # minimal_sum[i] will be the minimal sum of i different integers chosen among a set of possible integers
        self.minimal_sum = np.zeros(self.mark_nb - 1, dtype=int)
        # - a boolean array for marking already existing distances as used
        self.used_distance = np.empty(sum_first(self.mark_nb - 2) + 1, dtype=bool)
        # main constraints
        # dist_ij = mark_j - mark_i for j > i
        # mark_j = dist_0j for j > 0
        for i in range(1, mark_nb - 1):
            for j in range(i + 1, mark_nb):
                self.add_propagator(
                    (
                        [index(mark_nb, 0, j), index(mark_nb, 0, i), index(mark_nb, i, j)],
                        ALG_AFFINE_EQ,
                        [1, -1, -1, 0],
                    )
                )
        self.add_propagator((list(range(self.dist_nb)), ALG_ALLDIFFERENT, []))
        # redundant constraints
        for i in range(mark_nb - 1):
            for j in range(i + 1, mark_nb):
                if j - i < mark_nb - 1:
                    self.add_propagator(
                        (
                            [index(mark_nb, i, j), index(mark_nb, 0, mark_nb - 1)],
                            ALG_AFFINE_LEQ,
                            [1, -1, -sum_first(mark_nb - 1 - (j - i))],
                        ),
                        0,
                    )
        # break symmetries
        self.add_propagator(
            (
                [index(mark_nb, 0, 1), index(mark_nb, mark_nb - 2, mark_nb - 1)],
                ALG_AFFINE_LEQ,
                [1, -1, -1],
            ),
            0,
        )


def golomb_consistency_algorithm(statistics: NDArray, problem: GolombProblem) -> int:
    # first prune the search space
    ni_var_idx = first_not_instantiated_var_heuristic(problem.shr_domains_arr)  # no domains shared between vars
    if 1 < ni_var_idx < problem.mark_nb - 1:  # otherwise useless
        problem.used_distance.fill(False)
        # the following will mark at most sum(n-3) numbers as used
        # hence there will be at least n-2 unused numbers greater than 0
        for var_idx in range(index(problem.mark_nb, ni_var_idx - 2, ni_var_idx - 1) + 1):
            dist = problem.get_min_value(var_idx)
            if dist < len(problem.used_distance):
                problem.used_distance[dist] = True
        # let's compute the sum of non-used numbers
        distance = 1
        for j in range(0, problem.mark_nb - ni_var_idx):
            while problem.used_distance[distance]:
                distance += 1
            problem.minimal_sum[j + 1] = problem.minimal_sum[j] + distance
            distance += 1
        for i in range(ni_var_idx - 1, problem.mark_nb - 1):
            for j in range(i + 1, problem.mark_nb):
                var_idx = index(problem.mark_nb, i, j)
                new_min = problem.minimal_sum[j - i]
                # if new_min > self.get_max_value(var_idx):  # a bit slower
                #    return False
                problem.set_min_value(var_idx, new_min)
    # then apply BC
    return bound_consistency_algorithm(statistics, problem)



from pprint import pprint

from nucs.examples.donald_problem import DonaldProblem
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.heuristics import min_value_dom_heuristic, smallest_domain_var_heuristic
from nucs.statistics import STATS_SOLVER_SOLUTION_NB, get_statistics


class TestDonald:
    def test_donald(self) -> None:
        problem = DonaldProblem()
        solver = BacktrackSolver(
            problem, var_heuristic=smallest_domain_var_heuristic, dom_heuristic=min_value_dom_heuristic
        )
        solutions = solver.find_all()
        assert solver.statistics[STATS_SOLVER_SOLUTION_NB] == 1
        assert solutions[0] == [4, 3, 5, 9, 1, 8, 6, 2, 7, 0]


if __name__ == "__main__":
    problem = DonaldProblem()
    solver = BacktrackSolver(
        problem, var_heuristic=smallest_domain_var_heuristic, dom_heuristic=min_value_dom_heuristic
    )
    for solution in solver.solve():
        problem.pretty_print_solution(solution)
    pprint(get_statistics(solver.statistics))



from typing import Callable, Iterator, List, Optional

from nucs.problems.problem import Problem
from nucs.statistics import init_statistics


class Solver:
    """
    A solver.
    """

    def __init__(self, problem: Problem):
        self.problem = problem
        self.statistics = init_statistics()

    def solve(self) -> Iterator[List[int]]:  # type: ignore
        """
        Returns an iterator over the solutions.
        :return: an iterator
        """
        pass

    def solve_all(self, func: Optional[Callable] = None) -> None:
        """
        Finds all solutions.
        """
        for solution in self.solve():
            if func is not None:
                func(solution)

    def find_all(self) -> List[List[int]]:
        """
        Finds all solutions.
        """
        solutions = []
        self.solve_all(lambda solution: solutions.append(solution))
        return solutions

    def minimize(self, var_idx: int) -> Optional[List[int]]:  # type: ignore
        """
        Finds, if it exists, the solution to the problem that minimizes a given variable.
        :param variable_idx: the index of the variable
        :return: the solution if it exists or None
        """
        pass

    def maximize(self, var_idx: int) -> Optional[List[int]]:  # type: ignore
        """
        Finds, if it exists, the solution to the problem that maximizes a given variable.
        :param variable_idx: the index of the variable
        :return: the solution if it exists or None
        """
        pass



from pprint import pprint

from nucs.examples.knapsack_problem import KnapsackProblem
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.heuristics import first_not_instantiated_var_heuristic, max_value_dom_heuristic
from nucs.statistics import get_statistics


class TestKnapsack:
    def test_knapsack(self) -> None:
        problem = KnapsackProblem(
            [40, 40, 38, 38, 36, 36, 34, 34, 32, 32, 30, 30, 28, 28, 26, 26, 24, 24, 22, 22],
            [40, 40, 38, 38, 36, 36, 34, 34, 32, 32, 30, 30, 28, 28, 26, 26, 24, 24, 22, 22],
            55,
        )
        solver = BacktrackSolver(
            problem, var_heuristic=first_not_instantiated_var_heuristic, dom_heuristic=max_value_dom_heuristic
        )
        solution = solver.maximize(problem.weight)
        assert solution
        assert solution[problem.weight] == 54


if __name__ == "__main__":
    problem = KnapsackProblem(
        [40, 40, 38, 38, 36, 36, 34, 34, 32, 32, 30, 30, 28, 28, 26, 26, 24, 24, 22, 22],
        [40, 40, 38, 38, 36, 36, 34, 34, 32, 32, 30, 30, 28, 28, 26, 26, 24, 24, 22, 22],
        55,
    )
    solver = BacktrackSolver(
        problem, var_heuristic=first_not_instantiated_var_heuristic, dom_heuristic=max_value_dom_heuristic
    )
    solution = solver.maximize(problem.weight)
    pprint(get_statistics(solver.statistics))
    print(solution)



import os

from numba import types  # type: ignore
from numba.core import cgutils
from numba.experimental.function_type import _get_wrapper_address
from numba.extending import intrinsic


@intrinsic
def function_from_address(typingctx, func_type_ref, addr):  # type: ignore
    """
    Recovers a function from FunctionType and address.
    """
    func_type = func_type_ref.instance_type

    def codegen(context, builder, sig, args):  # type: ignore
        _, addr = args
        sfunc = cgutils.create_struct_proxy(func_type)(context, builder)
        llty = context.get_value_type(types.voidptr)
        addr_ptr = builder.inttoptr(addr, llty)
        sfunc.addr = addr_ptr
        return sfunc._getvalue()

    sig = func_type(func_type_ref, addr)
    return sig, codegen


def build_function_address_list(fcts, signature):  # type: ignore
    return [_get_wrapper_address(fct, signature) for fct in fcts]


NUMBA_DISABLE_JIT = os.getenv("NUMBA_DISABLE_JIT")



import argparse
from pprint import pprint

import pytest

from nucs.examples.quasigroup_problem import Quasigroup5Problem
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.heuristics import min_value_dom_heuristic, smallest_domain_var_heuristic
from nucs.statistics import STATS_SOLVER_SOLUTION_NB, get_statistics


class TestQuasigroup:
    @pytest.mark.parametrize(
        "size, solution_nb",
        [
            (7, 3),
            (8, 1),
            (9, 0),
            (10, 0),
            # (11, 5),
            # (12, 0),
        ],
    )
    def test_quasigroup5(self, size: int, solution_nb: int) -> None:
        problem = Quasigroup5Problem(size)
        solver = BacktrackSolver(
            problem, var_heuristic=smallest_domain_var_heuristic, dom_heuristic=min_value_dom_heuristic
        )
        solver.solve_all()
        assert solver.statistics[STATS_SOLVER_SOLUTION_NB] == solution_nb


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-n", type=int, default=10)
    args = parser.parse_args()
    problem = Quasigroup5Problem(args.n)
    solver = BacktrackSolver(
        problem, var_heuristic=smallest_domain_var_heuristic, dom_heuristic=min_value_dom_heuristic
    )
    solver.solve_all()
    pprint(get_statistics(solver.statistics))



from typing import Dict

import numpy as np
from numpy.typing import NDArray


def init_statistics() -> NDArray:
    """
    Inits a Numpy array for storing the statistics.
    :return: a Numpy array
    """
    return np.array([0] * STATS_MAX, dtype=np.int64)


def get_statistics(stats: NDArray) -> Dict[str, int]:
    """
    Returns the statistics as a dictionary.
    :param stats: a Numpy array of statistics
    :return: a dictionary
    """
    return {
        "OPTIMIZER_SOLUTION_NB": int(stats[STATS_OPTIMIZER_SOLUTION_NB]),
        "PROBLEM_FILTER_NB": int(stats[STATS_PROBLEM_FILTER_NB]),
        "PROBLEM_PROPAGATOR_NB": int(stats[STATS_PROBLEM_PROPAGATOR_NB]),
        "PROBLEM_VARIABLE_NB": int(stats[STATS_PROBLEM_VARIABLE_NB]),
        "PROPAGATOR_ENTAILMENT_NB": int(stats[STATS_PROPAGATOR_ENTAILMENT_NB]),
        "PROPAGATOR_FILTER_NB": int(stats[STATS_PROPAGATOR_FILTER_NB]),
        "PROPAGATOR_FILTER_NO_CHANGE_NB": int(stats[STATS_PROPAGATOR_FILTER_NO_CHANGE_NB]),
        "PROPAGATOR_INCONSISTENCY_NB": int(stats[STATS_PROPAGATOR_INCONSISTENCY_NB]),
        "SOLVER_BACKTRACK_NB": int(stats[STATS_SOLVER_BACKTRACK_NB]),
        "SOLVER_CHOICE_NB": int(stats[STATS_SOLVER_CHOICE_NB]),
        "SOLVER_CHOICE_DEPTH": int(stats[STATS_SOLVER_CHOICE_DEPTH]),
        "SOLVER_SOLUTION_NB": int(stats[STATS_SOLVER_SOLUTION_NB]),
    }


STATS_MAX = 12
(
    STATS_OPTIMIZER_SOLUTION_NB,
    STATS_PROBLEM_FILTER_NB,
    STATS_PROBLEM_PROPAGATOR_NB,
    STATS_PROBLEM_VARIABLE_NB,
    STATS_PROPAGATOR_ENTAILMENT_NB,
    STATS_PROPAGATOR_FILTER_NB,
    STATS_PROPAGATOR_FILTER_NO_CHANGE_NB,
    STATS_PROPAGATOR_INCONSISTENCY_NB,
    STATS_SOLVER_BACKTRACK_NB,
    STATS_SOLVER_CHOICE_NB,
    STATS_SOLVER_CHOICE_DEPTH,
    STATS_SOLVER_SOLUTION_NB,
) = tuple(range(STATS_MAX))



import numpy as np

from nucs.constants import PROP_CONSISTENCY
from nucs.numpy import new_data_by_values, new_shr_domains_by_values
from nucs.propagators.alldifferent_propagator import compute_domains_alldifferent


class TestAlldifferent:

    def test_compute_domains_1(self) -> None:
        domains = new_shr_domains_by_values([(3, 6), (3, 4), (2, 5), (2, 4), (3, 4), (1, 6)])
        data = new_data_by_values([])
        assert compute_domains_alldifferent(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[6, 6], [3, 4], [5, 5], [2, 2], [3, 4], [1, 1]]))

    def test_compute_domains_2(self) -> None:
        domains = new_shr_domains_by_values([(0, 0), (2, 2), (1, 2)])
        data = new_data_by_values([])
        assert compute_domains_alldifferent(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[0, 0], [2, 2], [1, 1]]))

    def test_compute_domains_3(self) -> None:
        domains = new_shr_domains_by_values([(0, 0), (0, 4), (0, 4), (0, 4), (0, 4)])
        data = new_data_by_values([])
        assert compute_domains_alldifferent(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[0, 0], [1, 4], [1, 4], [1, 4], [1, 4]]))



import sys

from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import MAX, MIN


@njit(cache=True)
def first_not_instantiated_var_heuristic(shr_domains: NDArray) -> int:
    """
    Chooses the first non-instantiated shared domain.
    :param shr_domains: the shared domains of the problem
    :return: the index of the shared domain
    """
    for dom_idx, shr_domain in enumerate(shr_domains):
        if shr_domain[MIN] < shr_domain[MAX]:
            return dom_idx
    return -1  # cannot happen


@njit(cache=True)
def last_not_instantiated_var_heuristic(shr_domains: NDArray) -> int:
    """
    Chooses the last non-instantiated shared domain.
    :param shr_domains: the shared domains of the problem
    :return: the index of the shared domain
    """
    for dom_idx in range(len(shr_domains) - 1, -1, -1):
        shr_domain = shr_domains[dom_idx]
        if shr_domain[MIN] < shr_domain[MAX]:
            return dom_idx
    return -1  # cannot happen


@njit(cache=True)
def smallest_domain_var_heuristic(shr_domains: NDArray) -> int:
    """
    Chooses the smallest shared domain and which is not instantiated.
    :param shr_domains: the shared domains of the problem
    :return: the index of the shared domain
    """
    min_size = sys.maxsize
    min_idx = -1
    for dom_idx, shr_domain in enumerate(shr_domains):
        size = shr_domain[MAX] - shr_domain[MIN]  # actually this is size - 1
        if 0 < size < min_size:
            min_idx = dom_idx
            min_size = size
    return min_idx


@njit(cache=True)
def greatest_domain_var_heuristic(shr_domains: NDArray) -> int:
    """
    Chooses the greatest shared domain and which is not instantiated.
    :param shr_domains: the shared domains of the problem
    :return: the index of the shared domain
    """
    max_size = 0
    max_idx = -1
    for dom_index, shr_domain in enumerate(shr_domains):
        size = shr_domain[MAX] - shr_domain[MIN]  # actually this is size - 1
        if max_size < size:
            max_idx = dom_index
            max_size = size
    return max_idx


@njit(cache=True)
def min_value_dom_heuristic(shr_domain: NDArray, shr_domain_copy: NDArray) -> int:
    """
    Chooses the first value of the domain.
    """
    value = shr_domain[MIN]
    shr_domain_copy[MIN] = value + 1
    shr_domain[MAX] = value
    return MAX


@njit(cache=True)
def max_value_dom_heuristic(shr_domain: NDArray, shr_domain_copy: NDArray) -> int:
    """
    Chooses the last value of the domain.
    """
    value = shr_domain[MAX]
    shr_domain_copy[MAX] = value - 1
    shr_domain[MIN] = value
    return MIN


@njit(cache=True)
def split_low_dom_heuristic(shr_domain: NDArray, shr_domain_copy: NDArray) -> int:
    """
    Chooses the first half of the domain.
    """
    value = (shr_domain[MIN] + shr_domain[MAX]) // 2
    shr_domain_copy[MIN] = value + 1
    shr_domain[MAX] = value
    return MAX



from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_AFFINE_EQ, ALG_COUNT_EQ


class MagicSequenceProblem(Problem):
    """
    Find a sequence x_0, ... x_n-1
    such that each x_i is the number of occurences of i in the sequence.

    CSPLIB problem #19 - https://www.csplib.org/Problems/prob019/
    """

    def __init__(self, n: int):
        self.n = n
        super().__init__([(0, n)] * n)
        for i in range(n):
            self.add_propagator((list(range(n)) + [i], ALG_COUNT_EQ, [i]))
        # redundant constraints
        self.add_propagator((list(range(n)), ALG_AFFINE_EQ, [1] * n + [n]))
        self.add_propagator((list(range(n)), ALG_AFFINE_EQ, list(range(n)) + [n]))



from typing import List

from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_AFFINE_EQ, ALG_ALLDIFFERENT

A, B, D, E, G, L, N, O, R, T = tuple(range(10))


class DonaldProblem(Problem):
    """
    This is the famous crypto-arithmetic puzzle DONALD+GERALD=ROBERT.
    The letters represent digits that have to be all different.
    """

    def __init__(self) -> None:
        super().__init__([(0, 9)] * 10)
        self.add_propagator(
            (
                [A, B, D, E, G, L, N, O, R, T],
                ALG_AFFINE_EQ,
                [200, -1000, 100002, 9900, 100000, 20, 1000, 0, -99010, -1, 0],
            )
        )
        self.add_propagator(
            ([A, B, D, E, G, L, N, O, R, T], ALG_ALLDIFFERENT, []),
        )

    def pretty_print_solution(self, solution: List[int]) -> None:
        print(
            {
                "A": solution[A],
                "B": solution[B],
                "D": solution[D],
                "E": solution[E],
                "G": solution[G],
                "L": solution[L],
                "N": solution[N],
                "O": solution[O],
                "R": solution[R],
                "T": solution[T],
            }
        )



from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_AFFINE_LEQ, ALG_EXACTLY_EQ, ALG_LEXICOGRAPHIC_LEQ


class SchurLemmaProblem(Problem):
    """
    CSPLIB problem #15 - https://www.csplib.org/Problems/prob015/
    """

    def __init__(self, n: int) -> None:
        super().__init__([(0, 1)] * n * 3)
        for x in range(n):
            self.add_propagator(([x * 3, x * 3 + 1, x * 3 + 2], ALG_EXACTLY_EQ, [1, 1]))
        for x in range(n):
            for y in range(n):
                z = (x + 1) + (y + 1) - 1
                if 0 <= z < n:
                    for k in range(3):
                        self.add_propagator(([3 * x + k, 3 * y + k, 3 * z + k], ALG_AFFINE_LEQ, [1, 1, 1, 2]))
        # breaking symmetries
        self.add_propagator(
            (list(range(0, n * 3, 3)) + list(range(1, n * 3, 3)) + list(range(2, n * 3, 3)), ALG_LEXICOGRAPHIC_LEQ, [])
        )



from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_ALLDIFFERENT
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.statistics import STATS_SOLVER_CHOICE_DEPTH, STATS_SOLVER_SOLUTION_NB


class TestBacktrackSolver:
    def test_solve_and_count(self) -> None:
        problem = Problem([(0, 99), (0, 99)])
        solver = BacktrackSolver(problem)
        for _ in solver.solve():
            pass
        assert solver.statistics[STATS_SOLVER_SOLUTION_NB] == 10000
        assert solver.statistics[STATS_SOLVER_CHOICE_DEPTH] == 2

    def test_solve(self) -> None:
        problem = Problem([(0, 1), (0, 1)])
        solver = BacktrackSolver(problem)
        solutions = [solution for solution in solver.solve()]
        assert len(solutions) == 4
        assert solutions[0] == [0, 0]
        assert solutions[1] == [0, 1]
        assert solutions[2] == [1, 0]
        assert solutions[3] == [1, 1]
        assert solver.statistics[STATS_SOLVER_SOLUTION_NB] == 4
        assert solver.statistics[STATS_SOLVER_CHOICE_DEPTH] == 2

    def test_solve_alldifferent(self) -> None:
        problem = Problem([(0, 2), (0, 2), (0, 2)])
        problem.add_propagator(([0, 1, 2], ALG_ALLDIFFERENT, []))
        solver = BacktrackSolver(problem)
        solutions = [solution for solution in solver.solve()]
        assert len(solutions) == 6
        assert solutions[0] == [0, 1, 2]
        assert solutions[1] == [0, 2, 1]
        assert solutions[2] == [1, 0, 2]
        assert solutions[3] == [1, 2, 0]
        assert solutions[4] == [2, 0, 1]
        assert solutions[5] == [2, 1, 0]
